/*****************************************************
Chip type               : ATmega8
AVR Core Clock frequency: 8,000000 MHz
*****************************************************/
#include <mega8.h>
#include <delay.h>

#define A PORTD.3
#define B PORTD.4                 
#define C PORTB.6
#define D PORTB.7
#define E PORTD.5
#define F PORTD.6
#define G PORTD.7
#define DP PORTB.0

#define A0 PORTD.0
#define A1 PORTD.1
#define A2 PORTD.2

#define R_upr 12000 //Ohms
#define R_lwr 2000   //Ohms
#define U_max_adc ((256*(R_upr+R_lwr))/R_lwr) //=U_max*100, 17,92V*100

#define U_bat_min ((1024 * 600)/U_max_adc)
#define U_bat_st1 ((1024 * 1000)/U_max_adc)
#define U_bat_st2 ((1024 * 1270)/U_max_adc)
#define U_bat_st3 ((1024 * 1470)/U_max_adc)
#define U_bat_max ((1024 * 1700)/U_max_adc)

#define psu_on do{PORTB.1 = 0;}while (0)
#define psu_off do{PORTB.1 = 1;}while (0)

#define charge_led PORTC.2
#define ready_led PORTC.3

#define ch_on  do {dsch_off; if (bs == st3) {charge_led = 1;} PORTC.1 = 1;}while(0)
#define ch_off  do {PORTC.1 = 0; if (bs == st3) {charge_led = 0;}}while(0)
#define dsch_on  do {ch_off; PORTC.4 = 1;}while(0)
#define dsch_off  do {PORTC.4 = 0;}while(0)

#define ADC_itn 16

volatile unsigned int ADC0_fr;
unsigned int volts;

volatile unsigned char display_buff [3], display_counter;

volatile bit adc_free, adc_busy, ttd, blink1s, blinks4, cdc;
volatile unsigned int tmr0, tmr1;

enum bat_states {st0, st1, st2, st3, st4, st5} bs;

void ports_init () 
{
 DDRB = 0b11111111; PORTB = 0b00000010;
 DDRC = 0b01111110; PORTC = 0b01000000;
 DDRD = 0b11111111; PORTD = 0b00000000;
}
void ADC_init ()
{
 MCUSR |= 1<<SE|1<<SM0; //adc noise reduction 
 ADCSR |= 1<<ADEN|1<<ADIE|1<<ADPS2|1<<ADPS1|1<<ADPS0; // div by 128
 ADMUX |= 1<<REFS1|1<<REFS0; //2,56 bg refs  
 ADCSRA |= (1<<ADSC);
}
void t0_init ()
{
 TCCR0 = 0; TCCR0 |= 1<<CS01|1<<CS00; // div by 64 _125khz
 TIMSK |= 1<<TOIE0;
}

void convert_data ()
{
 volts = (((unsigned long)ADC0_fr*U_max_adc)/10)>>10;
}
void clear_display () 
{ 
  A=1; B=1; C=1; D=1; E=1; F=1; G=1; DP=1; A0 = 0; A1 = 0; A2 = 0; 
}
void display_data ()
{
 (display_counter<2)?(display_counter++):(display_counter = 0);
 clear_display ();

 switch (display_buff [display_counter])
 {
  case 0: A=0; B=0; C=0; D=0; E=0; F=0; G=1; break;
  case 1: A=1; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 2: A=0; B=0; C=1; D=0; E=0; F=1; G=0; break;
  case 3: A=0; B=0; C=0; D=0; E=1; F=1; G=0; break;
  case 4: A=1; B=0; C=0; D=1; E=1; F=0; G=0; break;
  case 5: A=0; B=1; C=0; D=0; E=1; F=0; G=0; break;
  case 6: A=0; B=1; C=0; D=0; E=0; F=0; G=0; break;
  case 7: A=0; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 8: A=0; B=0; C=0; D=0; E=0; F=0; G=0; break;
  case 9: A=0; B=0; C=0; D=0; E=1; F=0; G=0; break; 
  case 10: A=1; B=1; C=1; D=0; E=0; F=0; G=1; break;     //L
  case 11: A=1; B=0; C=0; D=1; E=0; F=0; G=0; break;     //H
  case 12: A=1; B=1; C=1; D=1; E=1; F=1; G=1; break;     //
  default : A=1; B=1; C=1; D=1; E=1; F=1; G=0; break;    //-
   }
 
 switch (display_counter)
 {
  case 0: A0 = 1; break;
  case 1: A1 = 1; DP = 0; break;
  case 2: if (!display_buff [2]) {A2 = 0;} else {A2 = 1;} 
 } 
 }
void rnw_ddata ()
{
  char i;
  unsigned int volts_tmp;
  
  if (ttd)
  {
  volts_tmp = volts;  
  for (i=0; i<3; i++) {display_buff [i] = volts_tmp%10; volts_tmp /= 10;} 
  ttd = 0;
  }
}
void reed_adc ()
{
 while (!adc_free);
 adc_busy = 1; ADCSRA |= 1<<ADSC;
 while (adc_busy);
}

void bat_detect ();
void bat_ch_st1 ();
void bat_ch_st2 ();
void bat_ch_st3 ();
void bat_ch_st4 ();
void bat_ch_st5 ();

interrupt [ADC_INT] void adc_isr(void)
{
 static unsigned char n;
 static unsigned int ADC0_tmp;
   
 if ((adc_free)&&(!n)) {adc_free = 0; n = ADC_itn;}
 if ((!adc_free)&&(n))
 {
  ADC0_tmp  += ADCW;
  n--;
  ADCSRA |= 1<<ADSC;
  }
  else 
  {
  ADC0_fr = ADC0_tmp/ADC_itn;
  ADC0_tmp = 0; 
  adc_free = 1; adc_busy = 0; n = 0; 
  }
}
interrupt [TIM0_OVF] void timer0_ovf_isr(void)
{
 static unsigned int t0_counter, t0_counter1;
 
 if (t0_counter>488) {t0_counter = 0; ttd = 1; blink1s = ~blink1s;} 
 else {t0_counter++;}  
 if (t0_counter1>120) {t0_counter1 = 0; blinks4 = ~blinks4;} 
 else {t0_counter1++;}
 
 if (tmr0) {tmr0--;}
 if (tmr1) {tmr1--;} 
 
 display_data ();
}

void main(void)
{
ports_init ();
ADC_init ();
t0_init ();

bs = st0;

#pragma optsize-
WDTCR=0x1F;
WDTCR=0x0F;
#ifdef _OPTIMIZE_SIZE_
#pragma optsize+
#endif

#asm("sei")

while (1)
{    

        switch (bs)
       {
        case st0:  charge_led = blinks4; ready_led = ~blinks4; bat_detect (); break;
                 
        case st1: bat_ch_st1 (); charge_led = blinks4; ready_led = 0; break; 
        
        case st2:  bat_ch_st2 (); charge_led = blink1s; ready_led = 0; break;
        
        case st3: bat_ch_st3 ();  ready_led = 0; break;
        
        case st4: bat_ch_st4 (); charge_led = 0; ready_led = blink1s; break; 
        
        case st5: bat_ch_st5 (); charge_led = 0; ready_led = 1; break;     

        default: break;
       }
       
        
       rnw_ddata ();
       #asm("wdr")
      
}
}

void bat_detect ()
{
  ch_off; dsch_off; delay_ms (10); 
  reed_adc ();
 
    if ((ADC0_fr>U_bat_min)&&(ADC0_fr<U_bat_max)) 
    {
     if (ADC0_fr<U_bat_st1)         {bs = st1;}
     else if (ADC0_fr<U_bat_st2)  {bs = st2;}
     else if (ADC0_fr<U_bat_st3)  {bs = st3;}  
     else if (ADC0_fr>=U_bat_st3){bs = st4;}
     
     convert_data (); 
    }
    else {psu_off; bs = st0; convert_data ();}
 
}

void bat_ch_st1 ()
{
 static bit inop;
  
 if ((!tmr1)&&(!inop)) {inop = 1; psu_on; delay_ms (250); ch_on; tmr0 = 122;}
 if ((!tmr0)&&(inop)) 
 {inop = 0; psu_off; ch_off; tmr1 = 366; delay_ms (10); dsch_on; delay_ms (100); dsch_off;}  
 
 if (!inop) {bat_detect ();}
 else 
 {
    reed_adc ();
    if (ADC0_fr<(U_bat_min - 10)) {psu_off; ch_off; delay_ms (100); bat_detect ();}
 }
}
void bat_ch_st2 ()
{
 static bit inop;
  
 if (!inop) {inop = 1; psu_on; delay_ms (100); ch_on; tmr0 = 480;}
 if ((!tmr0)&&(inop)) 
 {inop = 0; psu_off; ch_off; delay_ms (10); dsch_on; delay_ms (100); dsch_off; delay_ms (10);}  
 
 if (!inop) {bat_detect ();}
 else 
 {
    reed_adc ();
    if (ADC0_fr<(U_bat_st1-10)) {psu_off; ch_off; delay_ms (100); bat_detect ();}
 }
}
void bat_ch_st3 ()
{
 static bit inop;
 static unsigned int ti, tp, tall;
 
 tall = 2000; 
 
 if ((!tmr1)&&(!inop)) 
 {
 inop = 1;  
 reed_adc ();
   {
    if (ADC0_fr < U_bat_st3) 
    {
    ti = tall*((U_bat_st3 - ADC0_fr)/(U_bat_st3 - U_bat_st2));
    if (ti<500) {ti = 500;}    
    if (ti>1750) {ti = 1750;}  
    tp = tall - ti;
    }
    else {ti = 100; tp = tall - ti;}
    bat_detect ();
   } 
 psu_on; delay_ms (100); ch_on; tmr0 = ti;} 
 
 if ((!tmr0)&&(inop)) 
 {inop = 0; if (tp>500) {psu_off;} ch_off; tmr1 = tp; delay_ms (10); dsch_on; delay_ms (100); dsch_off;}   

 cdc = 1;
}
void bat_ch_st4 ()
{
 static bit inop, rlock, llock;
 static char counter;
  
    if (cdc)
    {
       reed_adc ();
              
       llock = 1;
       if ((ADC0_fr<(U_bat_st2 + 20))&&(!rlock)) 
        {rlock = 1; if (counter<10) {counter++;} else {counter = 0; cdc = 0;}}
       if (ADC0_fr>(U_bat_st3 - 10)) {rlock = 0;}
        
       if (!rlock)
      { 
        if ((!tmr1)&&(!inop)) 
        {inop = 1; dsch_off; psu_off; convert_data (); delay_ms (10); 
         reed_adc ();
         ch_on; tmr0 = 100;}
        if ((!tmr0)&&(inop)) 
        {inop = 0; ch_off; tmr1 = 430; delay_ms (10); dsch_on; psu_on;}
      }
      else
      {
        if ((!tmr1)&&(!inop)) 
        {inop = 1; dsch_off; convert_data ();
         reed_adc (); 
         psu_on; delay_ms (100); ch_on; tmr0 = 400;}
        if ((!tmr0)&&(inop)) 
        {inop = 0; ch_off; tmr1 = 400; delay_ms (10); dsch_on; delay_ms (100); dsch_off;}      
      }
   }
    else if (llock) 
    {
        if (ADC0_fr>(U_bat_st3 - 10)) {llock = 0;}
        
        if ((!tmr1)&&(!inop)) 
        {inop = 1; dsch_off; convert_data (); 
        reed_adc (); 
        psu_on; delay_ms (100); ch_on; tmr0 = 400;}
        if ((!tmr0)&&(inop)) 
        {inop = 0; ch_off; tmr1 = 400; delay_ms (10); dsch_on; delay_ms (100); dsch_off;}
  }
    else if (!llock) 
    {
    psu_off; ch_off; dsch_off; 
    reed_adc (); 
    if (ADC0_fr>U_bat_st3){bs = st5;}
    else  {bat_detect ();}
    } 
    
    if ((ADC0_fr>U_bat_max)||(ADC0_fr<(U_bat_st2 - 30)))  {psu_off; bat_detect ();}
}
void bat_ch_st5 ()
{
        reed_adc ();
        if ((ADC0_fr < (U_bat_st2 - 20))||(ADC0_fr >(U_bat_st3 +20))) {bat_detect ();}
        else {convert_data ();}
}


     